@@ -12,18 +12,19 @@ import pathlib
 import subprocess
 import sys
 import typing
-from collections.abc import Iterator
+from contextlib import AbstractContextManager
 from contextlib import contextmanager
 from contextlib import nullcontext
 from functools import partial
 from subprocess import CompletedProcess
 from types import FunctionType
 from types import GenericAlias
-from typing import Any
-from typing import cast
-from typing import ContextManager
-from typing import NoReturn
 from typing import TYPE_CHECKING
+from typing import Any
+from typing import Callable
+from typing import NoReturn
+from typing import TypeVar
+from typing import cast
 
 import requests
 import rich
@@ -35,10 +36,19 @@ from ptscripts import process
 from ptscripts.virtualenv import VirtualEnv
 from ptscripts.virtualenv import VirtualEnvConfig
 
-if sys.version_info < (3, 11):
-    from typing_extensions import TypedDict, NotRequired
+if sys.version_info < (3, 10):
+    from typing_extensions import Concatenate
+    from typing_extensions import ParamSpec
 else:
-    from typing import TypedDict, NotRequired
+    from typing import Concatenate
+    from typing import ParamSpec
+
+if sys.version_info < (3, 11):
+    from typing_extensions import NotRequired
+    from typing_extensions import TypedDict
+else:
+    from typing import NotRequired
+    from typing import TypedDict
 
 try:
     import importlib.metadata
@@ -51,7 +61,15 @@ except ImportError:
 
 if TYPE_CHECKING:
     from argparse import ArgumentParser
+    from argparse import Namespace
     from argparse import _SubParsersAction
+    from collections.abc import Iterator
+
+
+Param = ParamSpec("Param")
+RetType = TypeVar("RetType")
+OriginalFunc = Callable[Param, RetType]
+DecoratedFunc = Callable[Concatenate[str, Param], RetType]
 
 log = logging.getLogger(__name__)
 
@@ -86,8 +104,10 @@ class Context:
     Context class passed to every command group function as the first argument.
     """
 
-    def __init__(self, parser: Parser):
+    def __init__(self, parser: Parser, debug: bool = False, quiet: bool = False) -> None:
         self.parser = parser
+        self._quiet = quiet
+        self._debug = debug
         self.repo_root = parser.repo_root
         theme = Theme(
             {
@@ -112,37 +132,39 @@ class Context:
         rich.reconfigure(stderr=True, **console_kwargs)
         self.venv = None
 
-    def print(self, *args, **kwargs):
+    def print(self, *args, **kwargs) -> None:
         """
         Print to stdout.
         """
         self.console_stdout.print(*args, **kwargs)
 
-    def debug(self, *args):
+    def debug(self, *args: str) -> None:
         """
         Print debug message to stderr.
         """
-        self.console.log(*args, style="log-debug", _stack_offset=2)
+        if self._debug:
+            self.console.log(*args, style="log-debug", _stack_offset=2)
 
-    def info(self, *args):
+    def info(self, *args: str) -> None:
         """
         Print info message to stderr.
         """
-        self.console.log(*args, style="log-info", _stack_offset=2)
+        if not self._quiet:
+            self.console.log(*args, style="log-info", _stack_offset=2)
 
-    def warn(self, *args):
+    def warn(self, *args: str) -> None:
         """
         Print warning message to stderr.
         """
         self.console.log(*args, style="log-warning", _stack_offset=2)
 
-    def error(self, *args):
+    def error(self, *args: str) -> None:
         """
         Print error message to stderr.
         """
         self.console.log(*args, style="log-error", _stack_offset=2)
 
-    def exit(self, status=0, message=None) -> NoReturn:  # type: ignore[misc]
+    def exit(self, status: int = 0, message: str | None = None) -> NoReturn:  # type: ignore[misc]
         """
         Exit the command execution.
         """
@@ -156,8 +178,8 @@ class Context:
 
     def _run(
         self,
-        *cmdline,
-        check=True,
+        *cmdline: str,
+        check: bool = True,
         timeout_secs: int | None = None,
         no_output_timeout_secs: int | None = None,
         capture: bool = False,
@@ -179,8 +201,8 @@ class Context:
 
     def run(
         self,
-        *cmdline,
-        check=True,
+        *cmdline: str,
+        check: bool = True,
         timeout_secs: int | None = None,
         no_output_timeout_secs: int | None = None,
         capture: bool = False,
@@ -237,10 +259,10 @@ class Context:
     @contextmanager
     def virtualenv(
         self,
-        name,
+        name: str,
         requirements: list[str] | None = None,
         requirements_files: list[pathlib.Path] | None = None,
-    ):
+    ) -> Iterator[VirtualEnv]:
         """
         Create and use a virtual environment.
         """
@@ -257,33 +279,32 @@ class Context:
         return requests.Session()
 
 
-class DefaultVirtualenvConfig:
+class DefaultVirtualEnv:
     """
     Simple class to hold registered imports.
     """
 
-    _instance: DefaultVirtualenvConfig | None = None
-    venv_config: VirtualEnvConfig
+    _instance: DefaultVirtualEnv | None = None
+    venv_config: VirtualEnvConfig | None
 
-    def __new__(cls):
+    def __new__(cls) -> DefaultVirtualEnv:
         """
         Method that instantiates a singleton class and returns it.
         """
         if cls._instance is None:
             instance = super().__new__(cls)
+            instance.venv_config = None
             cls._instance = instance
         return cls._instance
 
     @classmethod
-    def set_default_venv_config(cls, venv_config: VirtualEnvConfig) -> None:
+    def set_default_virtualenv_config(cls, venv_config: VirtualEnvConfig) -> None:
         """
-        Register an import.
+        Set the default tools requirements configuration.
         """
         instance = cls._instance
         if instance is None:
             instance = cls()
-        if venv_config and "name" not in venv_config:
-            venv_config["name"] = "default"
         instance.venv_config = venv_config
 
 
@@ -295,7 +316,7 @@ class RegisteredImports:
     _instance: RegisteredImports | None = None
     _registered_imports: dict[str, VirtualEnvConfig | None]
 
-    def __new__(cls):
+    def __new__(cls) -> RegisteredImports:
         """
         Method that instantiates a singleton class and returns it.
         """
@@ -316,7 +337,7 @@ class RegisteredImports:
         if import_module not in instance._registered_imports:
             instance._registered_imports[import_module] = venv_config
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[tuple[str, VirtualEnvConfig | None]]:
         """
         Return an iterator of all registered imports.
         """
@@ -334,14 +355,26 @@ class Parser:
     context: Context
     repo_root: pathlib.Path
 
-    def __new__(cls):
+    def __new__(cls) -> Parser:
         """
         Method that instantiates a singleton class and returns it.
         """
         if cls._instance is None:
+            # Let's do a litle manual parsing so that we can set debug or quiet early
+            debug = False
+            quiet = False
+            for arg in sys.argv[1:]:
+                if not arg.startswith("-"):
+                    break
+                if arg in ("-q", "--quiet"):
+                    quiet = True
+                    break
+                if arg in ("-d", "--debug"):
+                    debug = True
+                    break
             instance = super().__new__(cls)
             instance.repo_root = pathlib.Path.cwd()
-            instance.context = Context(instance)
+            instance.context = Context(instance, debug=debug, quiet=quiet)
             instance.parser = argparse.ArgumentParser(
                 prog="tools",
                 description="Python Tools Scripts",
@@ -410,16 +443,20 @@ class Parser:
             cls._instance = instance
         return cls._instance
 
-    def _process_registered_tool_modules(self):
-        default_venv: VirtualEnv | ContextManager[None]
-        default_venv_config = DefaultVirtualenvConfig().venv_config
+    def _process_registered_tool_modules(self) -> None:
+        default_venv: VirtualEnv | AbstractContextManager[None]
+        default_venv_config = DefaultVirtualEnv().venv_config
         if default_venv_config:
+            if "name" not in default_venv_config:
+                default_venv_config["name"] = "default"
+            default_venv_config["system_site_packages"] = True
+            default_venv_config["add_as_extra_site_packages"] = True
             default_venv = VirtualEnv(ctx=self.context, **default_venv_config)
         else:
             default_venv = nullcontext()
         with default_venv:
             for module_name, venv_config in RegisteredImports():
-                venv: VirtualEnv | ContextManager[None]
+                venv: VirtualEnv | AbstractContextManager[None]
                 if venv_config:
                     if "name" not in venv_config:
                         venv_config["name"] = module_name
@@ -435,7 +472,7 @@ class Parser:
                                 f"Could not import the registered tools module {module_name!r}: {exc}"
                             )
 
-    def parse_args(self):
+    def parse_args(self) -> None:
         """
         Parse CLI.
         """
@@ -443,12 +480,18 @@ class Parser:
         self._process_registered_tool_modules()
         options = self.parser.parse_args()
         if options.quiet:
+            self.context._quiet = True
+            self.context._debug = False
             logging.root.setLevel(logging.CRITICAL + 1)
         elif options.debug:
+            self.context._quiet = False
+            self.context._debug = True
             logging.root.setLevel(logging.DEBUG)
             self.context.console.log_path = True
             self.context.console_stdout.log_path = True
         else:
+            self.context._quiet = False
+            self.context._debug = False
             logging.root.setLevel(logging.INFO)
         if options.timestamps:
             for handler in logging.root.handlers:
@@ -459,13 +502,15 @@ class Parser:
         self.options = options
         if "func" not in options:
             self.context.exit(1, "No command was passed.")
-        log.debug(f"CLI parsed options {options}")
+        log.debug("CLI parsed options %s", options)
         options.func(options)
 
-    def __getattr__(self, attr):
+    def __getattr__(self, attr: str) -> Any:  # noqa: ANN401
         """
         Proxy unknown attributes to the parser instance.
         """
+        if attr == "options":
+            return self.__getattribute__(attr)
         return getattr(self.parser, attr)
 
 
@@ -480,7 +525,7 @@ class GroupReference:
     _instance: GroupReference | None = None
     _commands: dict[tuple[str, ...], CommandGroup]
 
-    def __new__(cls):
+    def __new__(cls) -> GroupReference:
         """
         Method that instantiates a singleton class and returns it.
         """
@@ -499,7 +544,7 @@ class GroupReference:
         if cli_name not in instance._commands:
             instance._commands[cli_name] = group
 
-    def __getitem__(self, item):
+    def __getitem__(self, item: tuple[str, ...]) -> CommandGroup:
         """
         Propogate getting a command parser to the underlying dict.
         """
@@ -511,7 +556,14 @@ class CommandGroup:
     Command group which holds the available tool functions.
     """
 
-    def __init__(self, name, help, description=None, parent=None, venv_config=None):
+    def __init__(
+        self,
+        name: str,
+        help: str,
+        description: str | None = None,
+        parent: Parser | CommandGroup | list[str] | tuple[str] | str | None = None,
+        venv_config: VirtualEnvConfig | None = None,
+    ) -> None:
         self.name = name
         if description is None:
             description = help
@@ -524,13 +576,15 @@ class CommandGroup:
             parent = [parent]
         if isinstance(parent, list):
             # NOTE: This means ordering of imports is important, but better than risking circular imports
-            GroupReference.add_command(tuple(parent + [name]), self)
+            GroupReference.add_command((*parent, name), self)
             parent = GroupReference()[tuple(parent)]
 
         if venv_config and "name" not in venv_config:
             venv_config["name"] = self.name
         self.venv_config = venv_config or {}
-        self.parser = parent.subparsers.add_parser(
+        if TYPE_CHECKING:
+            assert parent
+        self.parser = parent.subparsers.add_parser(  # type: ignore[union-attr, has-type]
             name.replace("_", "-"),
             help=help,
             description=description,
@@ -539,9 +593,9 @@ class CommandGroup:
             title="Commands",
             dest=f"{name.replace('-', '_')}_command",
         )
-        self.context = parent.context
+        self.context = parent.context  # type: ignore[union-attr, has-type]
 
-    def command(
+    def command(  # noqa: ANN201,C901,PLR0912,PLR0915
         self,
         func: FunctionType | None = None,
         *,
@@ -583,22 +637,23 @@ class CommandGroup:
 
         if arguments is None:
             if len(signature.parameters) > 1:
-                raise RuntimeError(
+                msg = (
                     f"'arguments' is a mandatory keyword argument to the '@{self.name}.command' "
                     "decorator when additional arguments(besides the required 'ctx' as first "
                     "argument) or keyword arguments are defined. Please update the decorated "
                     f"function {func_name!r} in {func_path!r}."
                 )
+                raise RuntimeError(msg)
             arguments = {}
 
         for key in arguments:
             if key not in signature.parameters:
-                raise RuntimeError(
-                    "Only pass argument names or keyword argument names on the 'arguments' "
-                    f"keyword for the '@{self.name}.command' decorated function {func_name!r} "
-                    f"in {func_path!r} which are also present in it's signature, {key!r} is "
-                    "not present."
+                msg = (
+                    "Only pass argument names or keyword argument names on the 'arguments' keyword "
+                    f"for the '@{self.name}.command' decorated function {func_name!r} in {func_path!r} "
+                    f"which are also present in it's signature, {key!r} is not present."
                 )
+                raise RuntimeError(msg)
 
         type_annotation = typing.get_type_hints(func)
         first_parameter_seen = False
@@ -606,10 +661,11 @@ class CommandGroup:
             if first_parameter_seen is False:
                 first_parameter_seen = True
                 if parameter.name != "ctx":
-                    raise RuntimeError(
+                    msg = (
                         f"'ctx' is a mandatory first argument to the '@{self.name}.command' "
                         f"decorated function {func_name!r} in {func_path!r}."
                     )
+                    raise RuntimeError(msg)
                 continue
             if parameter.annotation is parameter.empty:
                 # No typing annotations
@@ -631,13 +687,13 @@ class CommandGroup:
             if parameter.default is parameter.empty:
                 # Positional argument
                 kwargs["type"] = param_type
-                command.add_argument(parameter.name, **kwargs)
+                command.add_argument(parameter.name, **kwargs)  # type: ignore[arg-type]
                 continue
 
             if kwargs.get("nargs") == "*":
                 # Positional argument
                 kwargs["type"] = param_type
-                command.add_argument(parameter.name, **kwargs)
+                command.add_argument(parameter.name, **kwargs)  # type: ignore[arg-type]
                 continue
 
             # Keyword argument
@@ -655,26 +711,30 @@ class CommandGroup:
                         kwargs["action"] = action
 
             kwargs["default"] = parameter.default
-            if "help" in kwargs:
-                if parameter.default is not None:
-                    if not kwargs["help"].endswith("."):
-                        kwargs["help"] += "."
-                    kwargs["help"] += " [default: %(default)s]"
+            if "help" in kwargs and parameter.default is not None:
+                if not kwargs["help"].endswith("."):
+                    kwargs["help"] += "."
+                kwargs["help"] += " [default: %(default)s]"
             flags = kwargs.pop("flags", None)
             if flags is None:
                 flags = [f"--{parameter.name.replace('_', '-')}"]
             log.debug("Adding Command %r. Flags: %s; KwArgs: %s", name, flags, kwargs)
-            command.add_argument(*flags, **kwargs)
+            command.add_argument(*flags, **kwargs)  # type: ignore[arg-type]
         command.set_defaults(func=partial(self, func, venv_config=venv_config))
         return func
 
-    def __getattr__(self, attr):
+    def __getattr__(self, attr: str) -> Any:  # noqa: ANN401
         """
         Proxy unknown attributes to the parser instance.
         """
         return getattr(self.parser, attr)
 
-    def __call__(self, func, options, venv_config: VirtualEnvConfig | None = None):
+    def __call__(
+        self,
+        func: Callable[..., None],
+        options: Namespace,
+        venv_config: VirtualEnvConfig | None = None,
+    ) -> None:
         """
         Execute the selected tool function.
         """
@@ -692,22 +752,23 @@ class CommandGroup:
                     kwargs[name] = getattr(options, name)
 
         bound = signature.bind_partial(*args, **kwargs)
-        venv: VirtualEnv | ContextManager[None]
+        venv: VirtualEnv | None = None
         if venv_config:
             if "name" not in venv_config:
                 venv_config["name"] = getattr(options, f"{self.name}_command")
             venv = VirtualEnv(ctx=self.context, **venv_config)
         elif self.venv_config:
             venv = VirtualEnv(ctx=self.context, **self.venv_config)
+        if venv:
+            with venv:
+                previous_venv = self.context.venv
+                try:
+                    self.context.venv = venv
+                    func(self.context, *bound.args, **bound.kwargs)
+                finally:
+                    self.context.venv = previous_venv
         else:
-            venv = nullcontext()
-        with venv:
-            previous_venv = self.context.venv
-            try:
-                self.context.venv = venv
-                func(self.context, *bound.args, **bound.kwargs)
-            finally:
-                self.context.venv = previous_venv
+            func(self.context, *bound.args, **bound.kwargs)
 
 
 def command_group(
@@ -715,7 +776,7 @@ def command_group(
     help: str,
     description: str | None = None,
     venv_config: VirtualEnvConfig | None = None,
-    parent: CommandGroup | str | list[str] | tuple[str] | None = None,
+    parent: Parser | CommandGroup | list[str] | tuple[str] | str | None = None,
 ) -> CommandGroup:
     """
     Create a new command group.
